(* :Title: Package NFourierIntegral *)

(* :Context: Calculus`NFourierIntegral` *)

(* :Author: G.W. Inverarity, Met Office, London Road, Bracknell,
            RG12 2SZ, U.K. (gordon.inverarity@metoffice.com) *)

(* :Summary: Implement the algorithm of Bailey and Swarztrauber
  (1994), generalised to n dimensions and with asymmetric integration
  and output ranges (Inverarity, 2002), to evaluate integrals of the
  form I[f(t^(1), t^(2), ..., t^(n)), a_1, ..., a_n, b_1, ..., b_n; r,
  s](w^(1), w^(2), ..., w^(n)) = (Abs[s]/(2 Pi)^(1-r))^(n/2)
  Integrate[f[t^(1), t^(2), ... , t^(n)] Exp[i s (w^(1) t^(1) + w^(2)
  t^(2) + ... + w^(n) t^(n))], {t^(1), a_1, b_1}, {t^(2), a_2, b_2},
  ... , {t^(n), a_n, b_n}].  Call with NFourierIntegral[ex, intlim,
  translim, nint <, OPTIONS>]; where ex is the analytical expression
  to be integrated, intlim is the set of integration ranges for each
  integration variable entered as a list of lists consisting of the
  integration variable and two numeric limits (e.g., {{x, -1, 1}, {y,
  -1, 1}}), translim is the set of transformation variable limits over
  which the integral is to be tabulated (e.g., {{-1, 1}, {-1, 1}}) and
  nint is the list giving the number of intervals to be used in each
  dimension (e.g., {32, 64}).  There are seven options: Accuracy Goal
  (default value Infinity) and Precision Goal (default value
  Automatic) behave as for NIntegrate except that $MachinePrecision is
  used in place of WorkingPrecision, Compiled -> True (default)
  indicates that the integrand should be compiled prior to evaluation
  while False indicates that it should not, FourierParameters -> {r,
  s} provides a means to pass the parameters r and s to the routine
  (default values r=0, s=1), MinRecursion (default value 0) and
  MaxRecursion (default value 6) give the minimum and maximum number
  of recursive subdivisions, respectively, to be used when seeking
  convergence and Method -> Rectangular indicates that the rectangular
  rule should be used while Method -> Midpoint (default value)
  specifies the use of the midpoint rule. *)

(* :Package Version: 1.1 *)

(* :Mathematica Version: 4.1 *)

(* :Copyright: British Crown Copyright 2002 *)

(* :History: v1.0 20 February 2002 by G.W. Inverarity. 

             v1.1 1 May 2002 by G.W. Inverarity: 
                Context changed to Calculus`NFourierIntegral`.

                Usage message rewritten to be consistent with
                Inverarity (2002).

                "Midpoint" and "Rectangular" usage statements
                extended.

                Updated "System`Dump`$knownpackages" variable to allow
                "?NFourierIntegral" to produce a "More..." button.

                "General::*" errors changed to "NFourierIntegral::*".

                Backslashes used to improve readability of usage and
                error messages.

                Renamed the option "MaxIterations" to "MaxRecursion",
                changing its default value from 15 to 6, then added
                the option "MinRecursion", adapting the folding
                mechanism to take this into account, and the error
                message "NFourierIntegral::minmax" to make the routine
                more consistent with NIntegrate.

                Machine precision limitation acknowledged. 

                Simplified "NFourierIntegral::vars" and
                "NFourierIntegral::translim" checks.

                Swapped index order in "R" array to use the usual
                row/column indexing convention.

                "OuterProduct" altered to use "If" instead of
                "Which".
*)

(* :Keywords: Fourier integral, fast Fourier transform, FFT, numerical
    integration *)

(* :Sources:
  Bailey, D.H. and Swarztrauber, P.N., "A fast method for the
    numerical evaluation of continuous Fourier and Laplace
    transforms", SIAM J. Sci. Comput., 15, 1105-1110 (1994)

  Clenshaw, C.W. and Curtis, A.R., "A method for numerical integration
  on an automatic computer", Numerische Mathematik, 2, 197-205 (1960)

  Inverarity, G.W. "Fast computation of multi-dimensional Fourier
    integrals", SIAM J. Sci. Comput., in press (2002) 

  Press, W.H., Teukolsky, S.A., Vetterling, W.T. and Flannery, B.P.,
  "Numerical Recipes in FORTRAN: The Art of Scientific Computing", 2nd
  edition, Cambridge University Press, Cambridge (1992) *)

(* :Limitations: 
  The routine only works to machine precision.

  Array subtraction, but not addition, is quicker when written as a
  scalar product i.e., a - b = {1, -1} . {a, b}.  This yields a packed
  array when "a" and "b" are vectors but not when they are
  higher-dimensional arrays. *)

(* :Discussion: Given the ranges of finite t-values {{a_1, b_1}, ... ,
  {a_n, b_n}}, the output ranges of finite w-values {{c_1, d_1}, ... ,
  {c_n, d_n}} and the number of intervals to be used in each dimension
  {m_1, m_2, ... , m_n}, multiply all the c_l and d_l by s and proceed
  as for s = 1 without loss of generality if the final result is
  multiplied by |s|^(n/2).  Let beta_l = (b_l - a_l) / m_l, gamma_l =
  (d_l - c_l) / m_l and delta_l = beta_l gamma_l / 2 for l = 1, 2,
  ..., n.  Next, when using the extended midpoint rule, redefine the
  start points to be a'_l = a_l + beta_l / 2.  Now let t_j^(l) = a'_l
  + beta_l j and w_k^(l) = c_l + gamma_l k for 0 <= j, k <= m_l - 1
  and l = 1, 2, ..., n.  Define the array with dimensions 2 m_1 * 2
  m_2 * ... * 2 m_n whose general element is y[[j_1, ... , j_n]] =
  f[t_{j_1}^(1), t_{j_2}^(2), ..., t_{j_n}^(n)] Exp[i Sum[j_l (beta_l
  c_l + delta_l j_l), {l, 1, n}]] when 0 <= j_l <= m_l - 1 for all l =
  1, 2, ..., n, otherwise it is zero, and the sequence

  z[delta, m][[j]] = {Exp[-i delta j^2        , 0 <= j <= m - 1 
                     {Exp[-i delta (j - 2 m)^2, m <= j <= 2 m -1.

  Then I[f(t^(1), t^(2), ..., t^(n)), a_1, ..., a_n, b_1, ..., b_n; r,
  1](w_{k_1}^(1), w_{k_2}^(2), ..., w_{k_n}^(n)) = (2 Pi)^((r-1)n/2)
  Product[beta_l, {l, 1, n}] Exp[i Sum[a'_l w_{k_l}^(l) + delta_l
  k_l^2, {l, 1, n}]] InverseFourier[Table[Fourier[z[delta_1,
  m_1][[j_1]]] * Fourier[z[delta_2, m_2][[j_2]]] * ... *
  Fourier[z[delta_n, m_n][[j_n]]] * Fourier[y[[j_1, j_2, ..., j_n]]]]]
  when 0 <= k_l <= m_l - 1 for all l = 1, ..., n and using
  FourierParameters -> {1, -1} in the Fourier command.  Two separate
  adaptive implementations of Inverarity (2002) are used.  With the
  rectangular rule, the number of intervals is simply doubled and a
  convergence test applied.  When the midpoint rule is used, on the
  other hand, the number of intervals is again doubled but Richardson
  extrapolation (also known as extrapolation to the limit) is
  additionally employed.  Although interval tripling is more common
  when Richardson extrapolation is applied to the midpoint rule as it
  allows function evaluations to be reused in later iterations (Press
  et al., 1992, pp135-137), the additional time taken to evaluate the
  fast Fourier transforms of the resulting larger arrays nullifies
  this benefit as the number of iterations required to achieve
  convergence tends to be the same, on the whole.  Note that enhanced
  convergence only results when the integrand is finite at its
  endpoints otherwise the second Euler-Maclaurin formula, on which
  Richardson extrapolation depends, does not apply. *)

BeginPackage["Calculus`NFourierIntegral`"]

NFourierIntegral::usage = "NFourierIntegral[f, {{x_1, a_1, b_1}, ... \
, {x_n, a_n, b_n}}, {{c_1, d_1}, ... , {c_n, d_n}}, {m_1, ... , m_n}\
, FourierParameters \[Rule] {r, s}] gives a numerical approximation \
to the n-dimensional integral of f Exp[\[ImaginaryI] s (x_1 t_1 + ...\
 + x_n t_n)] (Abs[s]/ (2 \[Pi])^(1-r))^(n/2) with respect to x_1 from\
 a_1 to b_1, ... , x_n from a_n to b_n and generating output for t_1 \
from c_1 to d_1, ... , t_n from c_n to d_n in the form of an array of\
 dimensions m_1 * m_2 * ...* m_n." 

Midpoint::usage = "Midpoint is a setting for the Method option of \
NFourierIntegral.  Method -> Midpoint uses the midpoint quadrature rule."

Rectangular::usage = "Rectangular is a setting for the Method option \
of NFourierIntegral.  Method -> Rectangular uses the rectangular \
quadrature rule."

(* Ensure that this package appears in the list of known packages so
  that a "More..." button appears on entering "?NFourierIntegral". *)

If[!MemberQ[System`Dump`$knownpackages, "Calculus`NFourierIntegral`"],
  AppendTo[System`Dump`$knownpackages, "Calculus`NFourierIntegral`"]]

Begin["`Private`"]

(* Error messages. *)

NFourierIntegral::accg = NIntegrate::accg

NFourierIntegral::cv = "Prescribed convergence achieved after `1` \
iteration(s)." 

NFourierIntegral::dimensions = "The arrays of integration range(s) \
`1`, transform range(s) `2` and number(s) of intervals `3` have \
inconsistent lengths."

NFourierIntegral::fpopt = Fourier::fpopt

NFourierIntegral::fpreal = "Both values of the option \
FourierParameters -> `1` need to be real."

NFourierIntegral::intervals = "The number(s) of intervals `1` must \
be a positive integer or a list of positive integers."

NFourierIntegral::intvars = "The independent variables in `1` differ \
from those in the integration limits `2`."

NFourierIntegral::inum = "Integrand is not wholly numerical when \
evaluated at the integration nodes."

NFourierIntegral::minmax = NIntegrate::minmax

NFourierIntegral::nlim = "`1` = `2` is not a real limit of integration."

NFourierIntegral::nocv = "NFourierIntegral failed to converge after \
`1` iteration(s)."

NFourierIntegral::nonarray = "The integrand `1` should consist of a \
single expression."

NFourierIntegral::novar = "`1` cannot be used as the name of an \
integration variable." 

NFourierIntegral::optm = "Value of option Method -> `1` should be \
Rectangular or Midpoint."

NFourierIntegral::precg = "Value of option PrecisionGoal -> `1` is \
not Automatic, Infinity, or a positive integer or real number less \
than or equal to `2`." 

NFourierIntegral::translim = "Transform range specification in `1` \
does not consist of list(s) of the form {xmin, xmax} with xmin and \
xmax real numbers."

NFourierIntegral::vars = "Integration range specification `1` is \
neither a list nor list of lists of the form {x, xmin, xmax}."

(***************************** START **************************************)

(* Default options for NFourierIntegral. *)

Options[NFourierIntegral] = {AccuracyGoal -> Infinity, Compiled ->
  True, FourierParameters -> {0, 1}, MaxRecursion -> 6, MinRecursion
  -> 0, Method -> Midpoint, PrecisionGoal -> Automatic}

(* Apply the HoldAll attribute to enable erroneous calls to be
  returned unevaluated. *)

SetAttributes[NFourierIntegral, {HoldAll, ReadProtected}];


(* Wrapper for the main routine that returns the original call if an
   error occurs.  Turn off the Outer::normal message prior to calling
   nfourierintegral to inhibit warnings in OuterList and OuterProduct
   that are generated before compilation.  *)

NFourierIntegral[arg___] := Block[
  {result},               (* Result of calling nfourierintegral           *)
  Off[Outer::normal];
  result = nfourierintegral[arg];
  On[Outer::normal];
  result
  /; result =!= $Failed
]


(* Error if there are not exactly four non-option arguments in the
   function call. *)  

nfourierintegral[arg___] := $Failed /; !ArgumentCountQ[NFourierIntegral, 
  Length[Select[{arg}, !OptionQ[#] &]], 4, 4]


(* Error if not all the arguments after the fourth are options. *)

nfourierintegral[arg___] := With[{
  opts = Drop[{arg}, 4] (* Options                                        *)
  }, (
  Message[NFourierIntegral::nonopt, First @ Select[opts, !OptionQ[#] &], 4,
    HoldForm @ NFourierIntegral[arg]];
  $Failed
  ) /; Length[{arg}] > 4 && !VectorQ[opts, OptionQ]
];


(* Main routine. *)

nfourierintegral[
  ex_,                  (* Expression to be integrated                    *)
  range_,               (* Integration variable(s) and range(s)           *)
  trans_,               (* Transform variable range(s)                    *)
  nint_,                (* Number of intervals in each dimension          *)
  opts___?OptionQ       (* Options                                        *)
  ] := Block[{
  ag,                   (* Accuracy goal                                  *)
  alpha,                (* Dummy variable                                 *)
  ans,                  (* Result after iterating Inverarity2002          *)
  compile,              (* True means integrand compilation required      *)
  exvar,                (* Variables in ex                                *)
  f,                    (* Functional form of expression to be integrated *)
  fp,                   (* FourierParameters                              *)
  intlower,             (* Lower limit(s) of integration range(s)         *)
  intupper,             (* Upper limit(s) of integration range(s)         *)
  intvar,               (* Integration variable(s)                        *)
  j,                    (* Richardson extrapolation index                 *)
  k,                    (* Richardson extrapolation index                 *)
  lastconv,             (* Result of last convergence test                *)
  loopmax,              (* Maximum number of iterations                   *)
  loopmin,              (* Minimum number of iterations                   *)
  m,                    (* Number of intervals in each dimension          *)
  method,               (* "Rectangular" or "Midpoint" method             *)
  n,                    (* Number of dimensions                           *)
  pg,                   (* Precision goal                                 *)
  R,                    (* Richardson extrapolation array                 *)
  thisconv,             (* Result of this convergence test                *)
  translower,           (* Lower limit(s) of transform variable(s)        *)
  transupper            (* Upper limit(s) of transform variable(s)        *)
  },

(* Error for unrecognised options. *)

  alpha = Complement @@ 
    {Flatten[{opts}], Options[NFourierIntegral]}[[All, All, 1]];
  If[alpha =!= {},
    Message[NFourierIntegral::optx, alpha[[1]], 
      HoldForm @ NFourierIntegral[ex, range, trans, nint, opts]]; 
    Return[$Failed]
  ];

(* Extract parameter values. *)

  {ag, compile, fp, loopmax, loopmin, method, pg} = 
    Options[NFourierIntegral][[All, 1]] /. {opts} 
    /. Options[NFourierIntegral];

(* Error if "fp" does not have two elements. *)

  If[Length[fp] =!= 2,
    Message[NFourierIntegral::fpopt, fp];
    Return[$Failed]
  ];

(* Error if "fp" is not composed of real elements. *)

  If[!VectorQ[fp, RealQ],
    Message[NFourierIntegral::fpreal, fp];
    Return[$Failed]
  ];

(* Error if "compile" neither True nor False. *)

  If[!TrueQ @ Element[compile, Booleans],
    Message[NFourierIntegral::opttf, "Compiled", compile];
    Return[$Failed]
  ];

(* Error if "method" is neither "Rectangular" nor "Midpoint". *)

  If[!MemberQ[{Rectangular, Midpoint}, method],
    Message[NFourierIntegral::optm, method];
    Return[$Failed]
  ];

(* Error if "loopmin" is not a non-negative integer. *)

  If[Through @ Nand[IntegerQ, NonNegative][loopmin],
    Message[NFourierIntegral::par, MinRecursion, loopmin];
    Return[$Failed]
  ];

(* Error if "loopmax" is not a non-negative integer. *)

  If[Through @ Nand[IntegerQ, NonNegative][loopmax],
    Message[NFourierIntegral::par, MaxRecursion, loopmax];
    Return[$Failed]
  ];

(* Error if "loopmin" exceeds "loopmax". *)

  If[loopmin > loopmax,
    Message[NFourierIntegral::minmax, loopmin, loopmax];
    Return[$Failed]
  ];

(* Error if "ag" is not "Automatic", "Infinity" or a machine-sized
  real or integer. *)

  If[Nor[MemberQ[{Automatic, Infinity}, ag], 
    RealQ[ag] && MachineNumberQ @ N[ag]],
    Message[NFourierIntegral::accg, ag];
    Return[$Failed]
  ];
  ag = 10.0^(-If[ag === Automatic, $MachinePrecision - 10, ag]);

(* Error if "pg" is not "Automatic", "Infinity" or a positive integer
  or real number <= $MachinePrecision. *) 

  If[Nor[MemberQ[{Automatic, Infinity}, pg],
    RealQ[pg] && 0 < pg <= $MachinePrecision],
    Message[NFourierIntegral::precg, pg, $MachinePrecision];
    Return[$Failed]
  ];
  pg = 10.0^(-If[pg === Automatic, $MachinePrecision - 10, pg]);

(* Error if "ex" is an array. *)

  If[Dimensions[{ex}] =!= {1},
    Message[NFourierIntegral::nonarray, ex];
    Return[$Failed]
  ];

(* Error if "range" is neither a list consisting of 3 elements nor 
   a list of such lists. *) 

  alpha = ListList[range];
  If[!VectorQ[alpha, Length[#] === 3 &],
    Message[NFourierIntegral::vars, range];
    Return[$Failed]
  ];
  {intvar, intlower, intupper} = Transpose[alpha];

(* Error if any of the first integration range arguments is numeric. *)

  If[!VectorQ[intvar, !NumericQ[#] &],
    Message[NFourierIntegral::novar, First @ Select[intvar, NumericQ]]; 
    Return[$Failed]
  ];
  {intlower, intupper} = N[{intlower, intupper}];

(* Error if any of the second or third range arguments is not real.  
   alpha is the first list whose second or third entries are not
   real. *)

  If[!VectorQ[Join[intlower, intupper], RealQ],
    alpha = First @ Select[alpha, !MatchQ[#, {_, _?RealQ, _?RealQ}] &];
    Message[NFourierIntegral::nlim, alpha[[1]],
      First @ Select[Rest[alpha], !RealQ[#] &]]; 
    Return[$Failed]
  ];
        
(* Error if any of the transform variable ranges are not lists of two
   real numbers or lists of such lists. *)

  alpha = ListList[trans];
  If[!VectorQ[alpha, VectorQ[#, RealQ] && Length[#] === 2 &],
    Message[NFourierIntegral::translim, trans];
    Return[$Failed]
  ];

(* Scale the transform variable limits by fp[[2]]. *)

  {translower, transupper} = Transpose @ N[fp[[2]] alpha];

(* Error if the number(s) of intervals are not positive integers. *)

  m = Flatten[{nint}];
  If[!VectorQ[m, IntegerQ[#] && Positive[#] &],
    Message[NFourierIntegral::intervals, nint];
    Return[$Failed]
  ];

(* Error if the numbers of integration ranges, transformation variable
   ranges and intervals are not consistent. *)

  n = Length[m];
  If[!(Length[intlower] === Length[translower] === n), 
    Message[NFourierIntegral::dimensions, range, trans, nint];
    Return[$Failed]
  ];
    
(* Error if the variables used in the expression to be integrated do
   not form a subset of the variables in the integration ranges. *)

  exvar = Union @ Select[Level[{ex}, Depth[{ex}]], 
    AtomQ[#] && !NumericQ[#] &];
  If[Nand @@ (MemberQ[intvar, #] & /@ exvar),
    Message[NFourierIntegral::intvars, ex, range]; 
    Return[$Failed]
  ];

(* Extract the expression to be integrated and compile or not
   according to the value of "compile". *)

  f = If[compile, Compile, Function] @@ {intvar, ex};

(* Repeatedly double the number of intervals used to evaluate the
   integral until the result converges to the desired accuracy.
   Always check for two instances of convergence (cf. Clenshaw and
   Curtis (1960, p199)).  If the midpoint rule is being used then
   Richardson extrapolation can enhance the convergence.  The first
   column of the Richardson extrapolation table is based on
   Inverarity2002, doubling the number of intervals each time the
   second index increases by one but only returning "m" values each
   time.  Store table entries "R" for later reuse within this call
   only. *)

  R[j_, 0] := R[j, 0] = 
    With[{mm = m 2^j},           (* Number of intervals in this iteration *)
      Inverarity2002[mm][[Sequence @@ Range[1, mm, 2^j]]]
    ];

(* Later columns of the extrapolation table are computed iteratively
   based on the preceding column. *) 

  R[j_, k_] := R[j, k] = Developer`ToPackedArray[
    N[{4^k, -1} / (4^k - 1)] . {R[j + 1, k - 1], R[j, k - 1]}];

  ans = Catch @ If[method === Midpoint,

(* Evaluate the sequence R[0, i] for loopmin <= i <= loopmax, stopping
   when convergence has been achieved along the top row between R[0,
   i-1] and R[0, i] and down the penultimate column between R[0, i-1]
   and R[1, i-1]. *)

    Fold[(
      If[ConvergeQ[#1, R[0, #2]] && ConvergeQ[#1, R[1, #2 - 1]],
        Message[NFourierIntegral::cv, #2]; 
        Throw[{R[0, #2], $Done}]
      ];
      R[0, #2]) &, R[0, loopmin], Range[loopmin + 1, loopmax]],

(* For the rectangular rule, evaluate the sequence R[i, 0] for loopmin
   <= i <= loopmax, stopping when convergence has been achieved down
   the first column between R[i-2, 0] and R[i-1, 0] and between R[i-1,
   0] and R[i, 0]. *)

    lastconv = False;
    Fold[(
      thisconv = ConvergeQ[#1, R[#2, 0]];
      If[lastconv && thisconv,
        Message[NFourierIntegral::cv, #2];
        Throw[{R[#2, 0], $Done}],
        lastconv = thisconv
      ];
      R[#2, 0]) &, R[loopmin, 0], Range[loopmin + 1, loopmax]]
  ];

(* If no answer was thrown then return the best estimate along with an
   error message. *)

  If[FreeQ[ans, $Done],
    Message[NFourierIntegral::nocv, loopmax];
    ans,
    ans[[1]]
  ]
]

(***************************** STOP ***************************************)


(* :Title: Inverarity2002 *)

(* :Summary: Implement Eq. (4.8) of Inverarity (2002). *)

(* :History: v1.0 20 February 2002, by G.W. Inverarity. *)

(* :Discussion: Single application of Eq. (4.8) using number of
  subintervals contained in "m". *)

(***************************** START **************************************)

Inverarity2002[m_       (* Number of subintervals in each dimension       *)
              ] := Block[{
  alpha,                (* Dummy variable                                 *)
  beta,                 (* Physical resolution(s)                         *)
  delta,                (* Rescaled Fractional Fourier transform parameter*)
  gamma,                (* Transform variable resolution(s)               *)
  hi,                   (* Upper node(s) for function evaluation          *)
  lo,                   (* Lower node(s) for function evaluation          *)
  ysub,                 (* Nonzero unweighted elements of y               *)
  y                     (* Weighted function evaluations                  *)
},

(* Algorithm variables. *)

  beta = {1, -1} . {intupper, intlower} / m;
  gamma = {1, -1} . {transupper, translower} / m;
  delta = 0.5 beta gamma;

(* If any of the integration ranges have zero span then return an
   array of zeros. *)

  If[!VectorQ[beta, !Developer`ZeroQ[#] &],
    Return @ ZeroArray[m]];

(* Shift the integration limits by beta / 2 when using the
   second-order midpoint rule.  *)

  {lo, hi} = {intlower, {1, -1} . {intupper, beta}};
  If[method === Midpoint, {lo, hi} += 0.5 {beta, beta}];

(* Assign the upper-left subarray of "y" with the function evaluated
   at the nodes and scaled by the appropriate exponential factor,
   padding the rest of the array with complex zeros.  Exit the routine
   if "ysub" is not wholly numerical. *)

  ysub = If[n === 1,
           f /@ Flatten @ Range[lo, hi, beta],
           Developer`ToPackedArray @ Apply[f, Node[lo, hi, beta],{n}]
         ];
  alpha = Flatten[{ysub}];
  If[!VectorQ[alpha, NumericQ] || alpha === {},
    Message[NFourierIntegral::inum]; 
    Return[$Failed]
  ];

(* N.B. Using ZeroArray[m] as a third argument is quicker than using
   N[0. I] and returns a packed array, unlike N[0. I]. *)

  y = PadRight[
        ArrayMultiply[n][ysub, coeff1[beta translower, delta, m]], 
        m + m, ZeroArray[m]];

(* Compute, scale and return the circular convolution. *)
  
  ScalarMultiply[n][(Abs @ fp[[2]] / (2 Pi)^(1 - fp[[1]]))^(0.5 n) *
    Times @@ beta,  
    ArrayMultiply[n][Convolve[m], coeff2[lo translower, lo gamma, delta, m]]
  ]
]

(***************************** STOP ***************************************)


(* :Title: coeff1 *)

(* :Summary: Compute the weights applied to the function evaluations.
  Store values for later reuse. *)

(* :History: v1.0 22 February 2001, by G.W. Inverarity. *)

(* :Discussion: Use the formula derived in the theory.  Applying the N
  operator in the Array expression ensures large arrays are
  packed. *)

(***************************** START **************************************)

coeff1[alpha_,          (* Dummy variable                                 *)
       delta_,          (* Rescaled Fractional Fourier transform parameter*)
       m_               (* Number of intervals in each dimension          *)
      ] := coeff1[alpha, delta, m] =
  Exp @ Array[N[I Plus @@ ({##} (alpha + delta {##}))] &, m, 0]

(***************************** STOP ***************************************)


(* :Title: coeff2 *)

(* :Summary: Compute the weights applied to the circular convolution.
  Store values for later reuse. *)

(* :History: v1.0 20 February 2002, by G.W. Inverarity. *)

(* :Discussion: Use the formula derived in the theory.  Applying the N
  operator in the Array expression ensures large arrays are packed. *)

(***************************** START **************************************)

coeff2[alpha1_,         (* Dummy variable                                 *)
       alpha2_,         (* Dummy variable                                 *)
       delta_,          (* Rescaled Fractional Fourier transform parameter*)
       m_               (* Number of intervals in each dimension          *)
      ] := coeff2[alpha1, alpha2, delta, m] =
  Exp @ Array[N[I Plus @@ (alpha1 + {##}(alpha2 + delta {##}))] &, m, 0]

(***************************** STOP ***************************************)


(* :Title: z *)

(* :Summary: Compute the sequence with which the weighted function
  evaluations are convolved.  Store values for later reuse. *)

(* :History: v1.0 19 February 2002, by G.W. Inverarity. *)

(* :Limitations: Compilation errors occur when m >= 2^16. *)

(* :Discussion: Use the formula derived in the theory.  Applying the N
  operator ensures that large arrays are packed. *)

(***************************** START **************************************)

z[delta_,               (* Rescaled Fractional Fourier transform parameter*)
  m_                    (* Number of intervals (integer)                  *)
 ] := z[delta, m] = 
  With[{zz = ExpC[delta, m]}, (* First part of the "z" sequence           *)
    Fourier[Join[{1. + N[0. I]}, zz, Rest @ Reverse[zz]],
      FourierParameters -> {1, -1}]
  ]

(***************************** STOP ***************************************)


(* :Title: Convolve *)

(* :Summary: Compute the n-dimensional circular convolution employed
  in the theory. *)

(* :History: v1.0 19 February 2002, by G.W. Inverarity. *)

(* :Discussion: Evaluate the nested summations using Fast Fourier
  Transforms to perform the circular convolutions then reduce the
  padded array to the original dimensions of the problem.  N.B. This
  method is quicker than using Developer`FourierListConvolve. *)

(***************************** START **************************************)

Convolve = With[{
  fp = FourierParameters -> {1, -1} (* Fourier and InverseFourier option  *) 
  },
  InverseFourier[
    OuterProduct[n][Fourier[y, fp], Sequence @@ MapThread[z, {delta, #}]],
  fp][[Sequence @@ Range[#]]]
] &
  
(***************************** STOP ***************************************)


(* :Title: ArrayMultiply *)

(* :Summary: Elementwise multiply two complex arrays of equal rank. *)

(* :History: v1.0 6 September 2001, by G.W. Inverarity. *)

(* :Discussion: Compiled function. *)

(***************************** START **************************************)

ArrayMultiply = Compile[{{x, _Complex, #}, {y, _Complex, #}}, x y] &

(***************************** STOP ***************************************)


(* :Title: ConvergeQ *)

(* :Summary: Test whether the maximum absolute difference between
  the elements of two vectors is sufficiently small for convergence
  to be declared. *)

(* :History: v1.0 19 February 2002, by G.W. Inverarity. *)

(* :Discussion: The values pg and ag are inherited from nfourierintegral. *)

(***************************** START **************************************)

ConvergeQ = And @@ Negative @ Flatten @ Developer`ToPackedArray[
  {1, -pg} . {Abs[{1, -1} . {#1, #2}], Abs[#2]} - ag] &

(***************************** STOP ***************************************)


(* :Title: ExpC *)

(* :Summary: Evaluate an array of points around the unit circle in the
  complex plane. *)

(* :History: v1.0 19 February 2002, by G.W. Inverarity. *)

(* :Discussion: Compiled function. *)

(***************************** START **************************************)

ExpC = Compile[{x, {l, _Integer}}, Exp[-I x Range[l]^2]]

(***************************** STOP ***************************************)


(* :Title: ListList *)

(* :Summary: Turn a nested list or a single list into a list of
  lists. *)

(* :History: v1.0 6 September 2001, by G.W. Inverarity. *)

(* :Discussion: Flatten the input to the desired level. *)

(***************************** START **************************************)

ListList = Flatten[{{#}}, TensorRank[#]] &

(***************************** STOP ***************************************)


(* :Title: Node *)

(* :Summary: Construct a table of nodes in n dimensions.  Store values
  for later reuse. *)

(* :History: v1.0 19 February 2002, by G.W. Inverarity. *)

(* :Discussion: Take a Cartesian product of the sets of nodes in each
  spanning dimension generated by Range. *)

(***************************** START **************************************)

Node[x0_  ,             (* First node(s)                                  *)
     x1_,               (* Last node(s)                                   *)
     dx_                (* Node spacing(s)                                *)
      ] := Node[x0, x1, dx] = OuterList[n] @@ MapThread[Range, {x0, x1, dx}]

(***************************** STOP ***************************************)


(* :Title: OuterList *)

(* :Summary: Take the outer product of n 1-dimensional arrays. *)

(* :History: v1.0 20 February 2002, by G.W. Inverarity. *)

(* :Discussion: Compiled function with n arguments. *)

(***************************** START **************************************)

OuterList = 
  With[{vars = SymbolList[#]}, (* Dummy variable names                    *)
    Compile @@ 
      {Thread[{vars, _Real, 1}], Outer[List, Sequence @@ vars]}
  ] &

(***************************** STOP ***************************************)


(* :Title: OuterProduct *)

(* :Summary: Multiply a complex n-dimensional array by the outer
  product of n complex 1-dimensional arrays. *)

(* :History: v1.0 1 May 2002, by G.W. Inverarity. *)

(* :Discussion: Compiled function with n+1 arguments. *)

(***************************** START **************************************)

OuterProduct = If[# === 1, 
  ArrayMultiply[1],
  With[{vars = SymbolList[#]}, (* Dummy variable names                  *)
    Compile @@ 
      {Join[{{x, _Complex, #}}, Thread[{vars, _Complex, 1}]], 
        x Outer[Times, Sequence @@ vars]}
  ]
] &

(***************************** STOP ***************************************)


(* :Title: RealQ *)

(* :Summary: True if the argument is real. *)

(* :History: v1.0 15 February 2002, by G.W. Inverarity. *)

(* :Discussion: Convenient function. *)

(***************************** START **************************************)

RealQ = MemberQ[{Integer, Rational, Real}, Head[#]] &

(***************************** STOP ***************************************)


(* :Title: ScalarMultiply *)

(* :Summary: Multiply a complex array by a real scalar. *)

(* :History: v1.0 6 September 2001, by G.W. Inverarity. *)

(* :Discussion: Compiled function. *)

(***************************** START **************************************)

ScalarMultiply = Compile[{x, {y, _Complex, #}}, x y] &

(***************************** STOP ***************************************)


(* :Title: SymbolList *)

(* :Summary: Generate a list of symbols a1, a2, ..., an for use
  in compiled functions. *)

(* :History: v1.0 20 February 2002, by G.W. Inverarity. *)

(* :Discussion: Generate the list given the number of symbols. Store
  values for later reuse. *)

(***************************** START **************************************)

SymbolList[n_] := SymbolList[n] = Module[{a}, (* Variable prefix          *)
  Symbol[StringJoin @@ ToString /@ {a, #}] & /@ Range[n]]

(***************************** STOP ***************************************)


(* :Title: ZeroArray *)

(* :Summary: Create an array of complex zeros. *)

(* :History: v1.0 13 September 2001, by G.W. Inverarity. *)

(* :Discussion: Quicker than using Array[N[0. I] &, {dimensions}]. *)

(***************************** START **************************************)

ZeroArray = Developer`ToPackedArray @
  Fold[PadRight[{#1}, #2, {#1}] &, N[0. I], #] &

(***************************** STOP ***************************************)


End[];

Protect[NFourierIntegral];

EndPackage[];
